class WindowsDynamicError(Exception):
    pass


class AccessDeniedError(WindowsDynamicError):
    pass


class NoMoreEntriesError(WindowsDynamicError):
    pass


class HandleNotClosedSuccessfullyError(WindowsDynamicError):
    pass


class OpenProcessError(WindowsDynamicError):
    pass



import logging
import os
from pathlib import Path

log = logging.getLogger(__name__)

STREAM_FORMATTER = logging.Formatter("%(message)s")
FILE_FORMATTER = logging.Formatter("[%(asctime)s] [%(levelname)-5s] %(message)s")


class DelayedFileHandler(logging.FileHandler):
    def __init__(self, filename, *args, **kwargs):
        self.opened = False
        self._record_cache = []

        kwargs["delay"] = True
        logging.FileHandler.__init__(self, filename, *args, **kwargs)

    def set_filename(self, filename):
        if not self.opened:
            base_dir = os.path.dirname(self.baseFilename)
            self.baseFilename = os.fspath(os.path.join(base_dir, filename))
            self.flush_cache()
            self.opened = True

    def set_stream(self, stream):
        if not self.opened:
            self.stream = stream
            self.flush_cache()
            self.opened = True

    def flush_cache(self):
        if self._record_cache:
            for record in self._record_cache:
                logging.FileHandler.emit(self, record)
            self._record_cache = None

    def emit(self, record):
        if not self.opened:
            self._record_cache.append(record)
        else:
            logging.FileHandler.emit(self, record)

    def close(self):
        if not self.opened:
            # Close without being opened? Something probably broke
            if Path(self.baseFilename).parent.exists():
                log.info("Log written to file %s", Path(self.baseFilename).resolve())
                self.set_filename(self.baseFilename)
        logging.FileHandler.close(self)


def setup_logging(logger, path, verbosity, delay=False):
    if verbosity == 1:
        level = logging.ERROR
    elif verbosity == 2:
        level = logging.WARNING
    elif verbosity == 3:
        level = logging.INFO
    elif verbosity >= 4:
        level = logging.DEBUG
    else:
        level = logging.CRITICAL

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(STREAM_FORMATTER)
    stream_handler.setLevel(level)
    logger.addHandler(stream_handler)

    if path:
        file_handler = new_file_handler(path, delay)
        logger.addHandler(file_handler)

        root_logger = logging.getLogger()
        root_logger.addHandler(file_handler)
        root_logger.setLevel(logging.DEBUG)

    logger.setLevel(logging.DEBUG)


def reconfigure_log_file(logger, path, delay=False):
    file_handler = get_file_handler(logger)
    if file_handler is None:
        return

    # File handler is already opened, so close it and configure a new one
    if file_handler.stream:
        logger.removeHandler(file_handler)

        root_logger = logging.getLogger()
        root_logger.removeHandler(file_handler)

        file_handler.close()

        file_handler = new_file_handler(path, delay)
        logger.addHandler(file_handler)
        root_logger.addHandler(file_handler)
    else:
        file_handler.baseFilename = os.fspath(path)


def new_file_handler(path, delay=False):
    file_handler = DelayedFileHandler(path) if delay else logging.FileHandler(path)
    file_handler.setFormatter(FILE_FORMATTER)
    file_handler.setLevel(logging.DEBUG)
    return file_handler


def get_file_handler(logger):
    try:
        return next(handler for handler in logger.handlers if isinstance(handler, logging.FileHandler))
    except StopIteration:
        return None



from unittest.mock import Mock

import pytest
from dissect.target import Target
from dissect.target.filesystem import VirtualFile, VirtualFilesystem, VirtualSymlink


@pytest.fixture
def mock_file() -> Mock:
    return Mock()


@pytest.fixture
def mock_fs(mock_file) -> VirtualFilesystem:
    fs = VirtualFilesystem(case_sensitive=False)
    fs.makedirs("/foo/bar")
    fs.map_file_entry("/foo/bar/some-file", VirtualFile(fs, "some-file", mock_file))
    fs.map_file_entry("/foo/bar/own-file", VirtualFile(fs, "own-file", mock_file))
    fs.map_file_entry("/foo/bar/some-symlink", VirtualSymlink(fs, "some-symlink", "/foo/bar/some-file"))
    fs.map_file_entry("/foo/own-symlink", VirtualSymlink(fs, "own-symlink", "/foo/bar/own-file"))
    return fs


@pytest.fixture
def mock_target(mock_fs) -> Target:
    target = Target()
    target.fs.mount("/", mock_fs)
    target.filesystems.add(mock_fs)
    target.os = "mock"
    return target



from logging import getLogger
from typing import Iterable, Optional

from acquire.dynamic.windows.exceptions import AccessDeniedError
from acquire.dynamic.windows.handles import Handle, get_handles
from acquire.dynamic.windows.named_objects import NamedObject, NamedObjectType
from acquire.dynamic.windows.ntdll import (
    close_handle,
    open_directory_object,
    query_directory_object,
)

log = getLogger(__name__)


def collect_named_objects(path: str = "\\") -> list[NamedObject]:
    """Collects all named objects in the directory.

    Parameters:
        path: point to start searching from
    """

    try:
        dir_handle = open_directory_object(dir_name=path, root_handle=None)
    except AccessDeniedError:
        return []

    named_objects = query_directory_object(path_to_dir=path, dir_handle=dir_handle)

    dir_objects = [obj for obj in named_objects if obj.type_name == NamedObjectType.DIRECTORY]

    root_dir_path = path if path.endswith("\\") else f"{path}\\"

    for obj in dir_objects:
        named_objects += collect_named_objects(f"{root_dir_path}{obj.name}", dir_handle)

    close_handle(dir_handle)

    return named_objects


def collect_open_handles(handle_types: Optional[list[NamedObject]] = None) -> Iterable[Handle]:
    """Collect open handles

    Collect open handles and optionally provide a list to explicitly collect specific types of handles.

    Parameters:
        handle_types: list containing the handle types to collect as strings
    """
    for handle in get_handles():
        try:
            if not handle_types or NamedObjectType(handle.handle_type) in handle_types:
                yield handle
        # Continue if an invalid NamedObjectType is observed
        except ValueError:
            log.warning(f"Observed an unknown NamedObjectType: {handle.handle_type if handle else None}")
            continue



import hashlib
import io
from datetime import datetime, timezone

from dissect import cstruct

try:
    from Crypto.Cipher import AES, PKCS1_OAEP
    from Crypto.PublicKey import RSA
    from Crypto.Random import get_random_bytes

    HAS_PYCRYPTODOME = True
except ImportError:
    HAS_PYCRYPTODOME = False


c_acquire = cstruct.cstruct()
c_acquire.load(
    """
enum HeaderType : uint8 {
    PKCS1_OAEP = 0x1,
};

enum CipherType : uint8 {
    AES_256_GCM = 0x1,
};

struct file {
    char        magic[16];              // ENCRYPTEDACQUIRE
    uint8       version;                // Currently 1
    HeaderType  header_type;            // Currently PKCS1_OAEP
    uint16      header_size;            // Most often 512
    uint64      timestamp;              // Timestamp of write
    char        key_digest[32];         // SHA256(DER)
};

struct header {
    char        magic[12];              // KUSJESVANSRT
    CipherType  cipher_type;            // Currently AES_256_GCM
    uint8       key_length;             // Cipher key length
    uint8       iv_length;              // Cipher IV length
    uint8       _reserved;              // Reserved
    char        key[key_length];        // Cipher key
    char        iv[iv_length];          // Cipher IV
};

struct footer {
    char        magic[6];               // FOOTER
    uint16      length;                 // Digest length (precedes footer)
};
"""
)


FILE_MAGIC = b"ENCRYPTEDACQUIRE"
FILE_VERSION = 1
HEADER_MAGIC = b"KUSJESVANSRT"
FOOTER_MAGIC = b"FOOTER"


class EncryptedStream(io.RawIOBase):
    """Encrypted AES-256-GCM stream.

    Generates a random key and IV and uses AES-256-GCM to encrypt all written
    data.  The key and IV are encrypted with the given RSA public key and
    written as header.

    The header is included as AD to the AEAD cipher.
    The digest is written when the file is closed in the footer.

    Args:
        fh: The file-like object to write to.
        public_key: The RSA public key to encrypt the header with.
    """

    def __init__(self, fh, public_key):
        if not HAS_PYCRYPTODOME:
            raise ImportError("PyCryptodome is not available")

        self.fh = fh

        key = get_random_bytes(32)
        iv = get_random_bytes(12)
        self.cipher = AES.new(key, AES.MODE_GCM, nonce=iv)

        rsa = PKCS1_OAEP.new(RSA.import_key(public_key))

        plain_header = c_acquire.header(
            magic=HEADER_MAGIC,
            cipher_type=c_acquire.CipherType.AES_256_GCM,
            key_length=len(key),
            iv_length=len(iv),
            key=key,
            iv=iv,
        )
        sealed_header = rsa.encrypt(plain_header.dumps())

        file_header = c_acquire.file(
            magic=FILE_MAGIC,
            version=FILE_VERSION,
            header_type=c_acquire.HeaderType.PKCS1_OAEP,
            header_size=len(sealed_header),
            timestamp=int(datetime.now(timezone.utc).timestamp()),
            key_digest=key_fingerprint(rsa),
        )
        self.write_header(file_header.dumps() + sealed_header)

    def write_header(self, header):
        self.cipher.update(header)
        self.fh.write(header)

    def write(self, b):
        return self.fh.write(self.cipher.encrypt(b))

    def tell(self):
        return self.fh.tell()

    def seek(self, pos, whence=io.SEEK_CUR):
        raise TypeError("seeking is not allowed")

    def close(self):
        self.finalize()
        super().close()
        self.fh.close()

    def finalize(self):
        digest = self.cipher.digest()
        footer = c_acquire.footer(magic=FOOTER_MAGIC, length=len(digest))

        self.fh.write(digest + footer.dumps())
        self.fh.flush()
        if hasattr(self.cipher, "clean"):
            self.cipher.clean()


def key_fingerprint(pkey):
    if isinstance(pkey, PKCS1_OAEP.PKCS1OAEP_Cipher):
        pkey = pkey._key
    der = pkey.export_key("DER")

    return hashlib.sha256(der).digest()



from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from acquire.uploaders.minio import MinIO
from acquire.uploaders.plugin import upload_files_using_uploader
from acquire.uploaders.plugin_registry import UploaderRegistry


@pytest.fixture
def plugin_registry() -> UploaderRegistry:
    return UploaderRegistry("")


@pytest.fixture
def load_plugin(plugin_registry: UploaderRegistry) -> UploaderRegistry:
    plugin_registry.register("cloud", MinIO)
    return plugin_registry


@pytest.fixture
def minio_plugin(load_plugin):
    return load_plugin.get("cloud")


@pytest.fixture
def minio_instance(minio_plugin):
    return minio_plugin(upload={"endpoint": "test", "access_id": "test", "access_key": "test", "bucket": "test"})


@pytest.mark.parametrize(
    "arguments",
    [
        {"endpoint": "test", "access_id": "test", "access_key": "test", "bucket": "test"},
        {"endpoint": "test", "access_id": "test", "access_key": "test", "bucket": "test", "hello_world": "test"},
    ],
)
def test_minio_inputs(minio_plugin, arguments):
    minio = minio_plugin(upload=arguments)

    assert minio.endpoint == "test"
    assert minio.access_id == "test"
    assert minio.access_key == "test"
    assert minio.bucket_name == "test"


def test_minio_typeerror(minio_plugin):
    """Not enough arguments provided."""
    arguments = {}
    with pytest.raises(TypeError):
        minio_plugin(**arguments)


def test_minio_valueerror(minio_plugin):
    """Empty arguments."""
    arguments = {"endpoint": "", "access_id": "", "access_key": "", "bucket": ""}
    with pytest.raises(ValueError):
        minio_plugin(upload=arguments)


def test_upload_files(minio_instance):
    # Generates an internal error, which is caught
    upload_files_using_uploader(minio_instance, [Path("hello")])


def test_upload_file_multiple_failures(minio_instance: MinIO):
    mocked_upload = Mock()
    mocked_upload.side_effect = [Exception, Exception, Exception, "Hello"]
    minio_instance.upload_file = mocked_upload

    with patch("acquire.uploaders.plugin.log") as mocked_logger:
        # Only three retry attempts get made
        upload_files_using_uploader(minio_instance, [Path("hello")])
        mocked_logger.error.assert_called_with("Upload %s FAILED. See log file for details. Retrying", Path("hello"))

    # One extra gives a value error.
    mocked_upload.side_effect = [Exception, Exception, Exception, Exception]
    with patch("acquire.uploaders.plugin.log") as mocked_logger:
        upload_files_using_uploader(minio_instance, [Path("hello")])
        mocked_logger.error.assert_called_with("Upload %s FAILED after too many attempts. Stopping.", Path("hello"))



import errno
from unittest.mock import Mock, patch

import pytest
from dissect.target import Target
from dissect.target.exceptions import (
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
    SymlinkRecursionError,
)
from dissect.target.filesystem import VirtualFilesystem

from acquire.collector import Collector


def test_collector() -> None:
    target = Target("local")

    with patch("acquire.collector.log", autospec=True) as mock_log:
        fs_1 = VirtualFilesystem()
        fs_1.map_file("$MFT", None)
        target.fs.mount("C:", fs_1)
        target.filesystems.add(fs_1)

        fs_2 = VirtualFilesystem()
        fs_2.map_file("$MFT", None)
        target.fs.mount("D:", fs_2)
        target.filesystems.add(fs_2)

        collector = Collector(target, Mock())

        collector.collect_dir("C:", module_name="test")
        collector.collect_dir("D:", module_name="test")

        assert not mock_log.info.call_args.args[0] == "- Collecting file %s: Skipped (DEDUP)"


@pytest.fixture
def mock_collector(mock_target) -> Collector:
    collector = Collector(mock_target, Mock())
    return collector


MOCK_SEEN_PATHS = set()
MOCK_MODULE_NAME = "DUMMY"


def test_collector_collect_path_no_module_name(mock_collector) -> None:
    with pytest.raises(ValueError):
        mock_collector.collect_path("/some/path")


def test_collector_collect_path_dir_as_target_path(mock_target, mock_collector) -> None:
    with patch.object(mock_collector, "collect_dir", autospec=True):
        path = mock_target.fs.path("/foo/bar")
        mock_collector.collect_path(
            path,
            seen_paths=MOCK_SEEN_PATHS,
            module_name=MOCK_MODULE_NAME,
        )
        mock_collector.collect_dir.assert_called()


def test_collector_collect_path_dir(mock_collector) -> None:
    with patch.object(mock_collector, "collect_dir", autospec=True):
        mock_collector.collect_path(
            "/foo/bar",
            seen_paths=MOCK_SEEN_PATHS,
            module_name=MOCK_MODULE_NAME,
        )
        mock_collector.collect_dir.assert_called()


def test_collector_collect_path_file(mock_collector) -> None:
    with patch.object(mock_collector, "collect_file", autospec=True):
        mock_collector.collect_path(
            "/foo/bar/some-file",
            seen_paths=MOCK_SEEN_PATHS,
            module_name=MOCK_MODULE_NAME,
        )
        mock_collector.collect_file.assert_called()


def test_collector_collect_path_symlink(mock_collector) -> None:
    with patch.object(mock_collector, "collect_symlink", autospec=True):
        with patch.object(mock_collector, "collect_file", autospec=True):
            mock_collector.collect_path(
                "/foo/bar/some-symlink",
                follow=False,
                seen_paths=MOCK_SEEN_PATHS,
                module_name=MOCK_MODULE_NAME,
            )
            mock_collector.collect_symlink.assert_called()
            mock_collector.collect_file.assert_not_called()


def test_collector_collect_path_symlink_follow(mock_collector) -> None:
    with patch.object(mock_collector, "collect_symlink", autospec=True):
        with patch.object(mock_collector, "collect_file", autospec=True):
            mock_collector.collect_path(
                "/foo/bar/some-symlink",
                follow=True,
                seen_paths=MOCK_SEEN_PATHS,
                module_name=MOCK_MODULE_NAME,
            )
            mock_collector.collect_symlink.assert_called()
            mock_collector.collect_file.assert_called()


@pytest.mark.parametrize(
    "path, symlink_called, file_called",
    [
        (
            "/foo/bar/own-file",
            False,
            False,
        ),
        (
            "/foo/own-symlink",
            True,
            False,
        ),
        (
            "/foo/bar/some-file",
            False,
            True,
        ),
        (
            "/foo/bar/some-symlink",
            True,
            True,
        ),
    ],
)
def test_collector_collect_path_skip_list(mock_collector, path, symlink_called, file_called) -> None:
    with patch.object(mock_collector, "skip_list", new={"/foo/bar/own-file"}):
        with patch.object(mock_collector, "collect_symlink", autospec=True):
            with patch.object(mock_collector, "collect_file", autospec=True):
                mock_collector.collect_path(
                    path,
                    follow=True,
                    seen_paths=MOCK_SEEN_PATHS,
                    module_name=MOCK_MODULE_NAME,
                )
                if symlink_called:
                    mock_collector.collect_symlink.assert_called()
                else:
                    mock_collector.collect_symlink.assert_not_called()

                if file_called:
                    mock_collector.collect_file.assert_called()
                else:
                    mock_collector.collect_file.assert_not_called()


def test_collector_collect_glob(mock_collector) -> None:
    with patch.object(mock_collector, "collect_file", autospec=True):
        mock_collector.collect_glob(
            "/foo/bar/*",
            module_name=MOCK_MODULE_NAME,
        )
        assert len(mock_collector.collect_file.mock_calls) == 3
        assert mock_collector.collect_file.call_args.kwargs.get("module_name", None) == MOCK_MODULE_NAME


def test_collector_collect_path_non_existing_file(mock_collector) -> None:
    with patch("acquire.collector.log", autospec=True) as mock_log:
        with patch.object(mock_collector, "report", autospec=True) as mock_report:
            mock_collector.collect_path(
                "/foo/bar/non-existing-file",
                seen_paths=MOCK_SEEN_PATHS,
                module_name=MOCK_MODULE_NAME,
            )
            mock_report.add_path_missing.assert_called()
            mock_log.error.assert_called()
            assert mock_log.error.call_args.args[0] == "- Path %s is not found"


def test_collector_collect_path_no_file_type(mock_target, mock_collector) -> None:
    path = mock_target.fs.path("/foo/bar/non-existing-file")
    with patch("acquire.collector.log", autospec=True) as mock_log:
        with patch.object(mock_collector, "report", autospec=True) as mock_report:
            with patch.object(path, "get", return_value=True, autospec=True):
                with patch.object(path, "is_dir", return_value=False, autospec=True):
                    with patch.object(path, "is_file", return_value=False, autospec=True):
                        with patch.object(path, "is_symlink", return_value=False, autospec=True):
                            mock_collector.collect_path(
                                path,
                                seen_paths=MOCK_SEEN_PATHS,
                                module_name=MOCK_MODULE_NAME,
                            )
                            mock_report.add_path_failed.assert_called()
                            mock_log.error.assert_called()
                            assert mock_log.error.call_args.args[0] == "- Don't know how to collect %s in module %s"


@pytest.mark.parametrize(
    "report_func, exception, log_msg",
    [
        (
            "add_path_missing",
            OSError(errno.ENOENT, "foo"),
            "- Path %s is not found",
        ),
        (
            "add_path_failed",
            OSError(errno.EACCES, "foo"),
            "- Permission denied while accessing path %s",
        ),
        (
            "add_path_failed",
            OSError(255, "foo"),
            "- OSError while collecting path %s",
        ),
        (
            "add_path_missing",
            FileNotFoundError,
            "- Path %s is not found",
        ),
        (
            "add_path_missing",
            NotADirectoryError,
            "- Path %s is not found",
        ),
        (
            "add_path_missing",
            NotASymlinkError,
            "- Path %s is not found",
        ),
        (
            "add_path_missing",
            SymlinkRecursionError,
            "- Path %s is not found",
        ),
        (
            "add_path_missing",
            ValueError,
            "- Path %s is not found",
        ),
        (
            "add_path_failed",
            Exception,
            "- Failed to collect path %s",
        ),
    ],
)
def test_collector_collect_path_with_exception(mock_target, mock_collector, report_func, exception, log_msg) -> None:
    path = mock_target.fs.path("/foo/bar/non-existing-file")
    with patch("acquire.collector.log", autospec=True) as mock_log:
        with patch.object(mock_collector, "report", autospec=True) as mock_report:
            with patch.object(path, "get", side_effect=exception, autospec=True):
                mock_collector.collect_path(
                    path,
                    seen_paths=MOCK_SEEN_PATHS,
                    module_name=MOCK_MODULE_NAME,
                )
                report_func = getattr(mock_report, report_func)
                report_func.assert_called()
                mock_log.error.assert_called()
                assert mock_log.error.call_args.args[0] == log_msg



import io
from pathlib import Path
from typing import BinaryIO, Optional, Union

from dissect.target import Target
from dissect.target.filesystem import FilesystemEntry

import acquire.utils


class Output:
    """Base class to implement acquire output formats with.

    New output formats must sub-class this class.

    Args:
        target: The target that we're using acquire on.
    """

    def init(self, target: Target):
        pass

    def write(
        self,
        output_path: str,
        fh: BinaryIO,
        entry: Optional[Union[FilesystemEntry, Path]],
        size: Optional[int] = None,
    ) -> None:
        """Write a filesystem entry or file-like object to the implemented output type.

        Args:
            output_path: The path of the entry in the output format.
            fh: The file-like object of the entry to write.
            entry: The optional filesystem entry of the entry to write.
            size: The optional file size in bytes of the entry to write.
        """
        raise NotImplementedError()

    def write_entry(
        self,
        output_path: str,
        entry: Optional[Union[FilesystemEntry, Path]],
        size: Optional[int] = None,
    ) -> None:
        """Write a filesystem entry to the output format.

        Args:
            output_path: The path of the entry in the output format.
            entry: The optional filesystem entry of the entry to write.
            size: The optional file size in bytes of the entry to write.
        """
        with entry.open() as fh:
            self.write(output_path, fh, entry, size)

    def write_bytes(
        self,
        output_path: str,
        data: bytes,
        entry: Optional[Union[FilesystemEntry, Path]] = None,
        size: Optional[int] = None,
    ) -> None:
        """Write raw bytes to the output format.

        Args:
            output_path: The path of the entry in the output format.
            data: The raw bytes to write.
            entry: The optional filesystem entry of the entry to write.
            size: The optional file size in bytes of the entry to write.
        """

        stream = io.BytesIO(data)
        self.write(output_path, stream, entry, size)

    def write_volatile(
        self,
        output_path: str,
        entry: Optional[Union[FilesystemEntry, Path]],
        size: Optional[int] = None,
    ) -> None:
        """Write specified path to the output format.
        Handles files that live in volatile filesystems. Such as procfs and sysfs.

        Args:
            output_path: The path of the entry in the output format.
            entry: The optional filesystem entry of the entry to write.
            size: The optional file size in bytes of the entry to write.
        """
        try:
            fh = acquire.utils.VolatileStream(Path(entry.path))
            buf = fh.read()
            size = size or len(buf)
        except (OSError, PermissionError):
            # Various OSErrors can occur here.
            # If one does occur, we'd still like to have the corresponding entry.
            buf = b""
            size = 0

        self.write_bytes(output_path, buf, entry, size)

    def close(self) -> None:
        """Closes the output."""
        raise NotImplementedError()



from __future__ import annotations

import ctypes
from enum import IntEnum, IntFlag
from typing import List

from acquire.dynamic.windows.exceptions import (
    AccessDeniedError,
    HandleNotClosedSuccessfullyError,
    NoMoreEntriesError,
)
from acquire.dynamic.windows.named_objects import NamedObject
from acquire.dynamic.windows.types import (
    BOOL,
    DWORD,
    HANDLE,
    IO_STATUS_BLOCK,
    LPVOID,
    NTSTATUS,
    NULL,
    OBJECT_DIRECTORY_INFORMATION,
    PHANDLE,
    PULONG,
    PUNICODE_STRING,
    PVOID,
    ULONG,
    UNICODE_STRING,
)

# https://github.com/pentestmonkey/windows-privesc-check/blob/master/wpc/ntobj.py


ntdll = ctypes.windll.ntdll
NtQueryInformationFile = ntdll.NtQueryInformationFile
NtQueryInformationFile.argtypes = [
    HANDLE,
    ctypes.POINTER(IO_STATUS_BLOCK),
    LPVOID,
    ULONG,
    DWORD,
]
ntdll.NtQueryInformationFile.restype = NTSTATUS

NtQuerySystemInformation = ntdll.NtQuerySystemInformation
NtQuerySystemInformation.argtypes = [
    ULONG,
    LPVOID,
    DWORD,
    ctypes.POINTER(DWORD),
]
ntdll.NtQuerySystemInformation.restype = NTSTATUS

NtQueryObject = ntdll.NtQueryObject
NtQueryObject.argtypes = [
    HANDLE,
    ULONG,
    LPVOID,
    DWORD,
    PULONG,
]
NtQueryObject.restype = NTSTATUS

STANDARD_RIGHTS_ALL = 0x001F0000
BUFFER_SIZE = 1024


class OBJECT_ATTRIBUTES(ctypes.Structure):
    _fields_ = [
        ("Length", ULONG),
        ("RootDirectory", HANDLE),
        ("ObjectName", PUNICODE_STRING),
        ("Attributes", ULONG),
        ("SecurityDescriptor", PVOID),
        ("SecurityQualityOfService", PVOID),
    ]


class NtStatusCode(IntEnum):
    STATUS_SUCCESS = 0x00000000
    STATUS_MORE_ENTRIES = 0x00000105
    STATUS_ACCESS_DENIED = 0xC0000022
    STATUS_INFO_LENGTH_MISMATCH = 0xC0000004
    STATUS_INVALID_HANDLE = 0xC0000008
    STATUS_NO_MORE_ENTRIES = 0x8000001A
    STATUS_BUFFER_OVERFLOW = 0x80000005


class ACCESS_MASK(IntFlag):
    DIRECTORY_QUERY = 0x0001
    DIRECTORY_TRAVERSE = 0x0002
    DIRECTORY_CREATE_OBJECT = 0x0004
    DIRECTORY_CREATE_SUBDIRECTORY = 0x0008
    DIRECTORY_ALL_ACCESS = STANDARD_RIGHTS_ALL | 0xF


class OBJ_ATTR(IntFlag):
    """
    https://github.com/tpn/winsdk-10/blob/master/Include/10.0.10240.0/shared/ntdef.h
    https://docs.microsoft.com/en-us/windows-hardware/drivers/kernel/object-handles
    """

    OBJ_INHERIT = 0x00000002
    OBJ_PERMANENT = 0x00000010
    OBJ_EXCLUSIVE = 0x00000020
    OBJ_CASE_INSENSITIVE = 0x00000040
    OBJ_OPENIF = 0x00000080
    OBJ_OPENLINK = 0x00000100
    OBJ_KERNEL_HANDLE = 0x00000200
    OBJ_FORCE_ACCESS_CHECK = 0x00000400
    OBJ_IGNORE_IMPERSONATED_DEVICEMAP = 0x00000800
    OBJ_VALID_ATTRIBUTES = 0x00000FF2


# Define interface
NtOpenDirectoryObject = ntdll.NtOpenDirectoryObject
NtOpenDirectoryObject.argtypes = [
    PHANDLE,  # DirectoryHandle
    DWORD,  # DesiredAccess
    ctypes.POINTER(OBJECT_ATTRIBUTES),  # ObjectAttributes
]
NtOpenDirectoryObject.restype = NTSTATUS


# Define interface
NtQueryDirectoryObject = ntdll.NtQueryDirectoryObject
NtQueryDirectoryObject.argtypes = [
    HANDLE,  # DirectoryHandle
    PVOID,  # Buffer
    ULONG,  # Length
    BOOL,  # ReturnSingleEntry
    BOOL,  # RestartScan
    PULONG,  # Context
    PULONG,  # ReturnLength
]
NtQueryDirectoryObject.restype = NTSTATUS

RtlNtStatusToDosError = ntdll.RtlNtStatusToDosError
RtlNtStatusToDosError.argtypes = [NTSTATUS]
RtlNtStatusToDosError.restype = ULONG

CloseHandle = ctypes.windll.kernel32.CloseHandle
CloseHandle.argtypes = [HANDLE]
CloseHandle.restype = BOOL


def initialize_object_attributes(
    destination_attributes: OBJECT_ATTRIBUTES,
    name: PUNICODE_STRING,
    attributes: OBJ_ATTR,
    root_directory: HANDLE,
    security_descriptor: PVOID,
) -> None:
    """Initializes the OBJECT_ATTRIBUTES structure.

    Allocates said information at the address of InitializedAttributes
    """
    ctypes.memset(ctypes.addressof(destination_attributes), 0, ctypes.sizeof(destination_attributes))
    destination_attributes.Length = ctypes.sizeof(destination_attributes)
    destination_attributes.ObjectName = name
    destination_attributes.Attributes = attributes
    destination_attributes.RootDirectory = root_directory
    destination_attributes.SecurityDescriptor = security_descriptor
    destination_attributes.SecurityQualityOfService = None


def close_handle(handle: HANDLE) -> None:
    """Closes an opened handle."""
    if not CloseHandle(handle):
        raise HandleNotClosedSuccessfullyError()


def validate_ntstatus(status: NTSTATUS) -> None:
    """Validates the result status of a Nt call

    Parameters:
        status: the return value of a ntcall
    """
    if status == NtStatusCode.STATUS_ACCESS_DENIED:
        raise AccessDeniedError("Couldn't access the resource")
    if status == NtStatusCode.STATUS_NO_MORE_ENTRIES:
        raise NoMoreEntriesError("There are no more entries available.")
    if status not in (NtStatusCode.STATUS_SUCCESS, NtStatusCode.STATUS_MORE_ENTRIES):
        raise ctypes.WinError(RtlNtStatusToDosError(status))


def open_directory_object(dir_name: str, root_handle: HANDLE = None) -> HANDLE:
    """Opens a handle to a specific directory structure of NamedObjects.

    Parameters:
        dir_name: Specific directory we want to try and open.
        root_handle: From which point we want to start querying the object.
    """

    object_name = UNICODE_STRING.from_str(dir_name)
    p_name = ctypes.pointer(object_name)

    obj_attrs = OBJECT_ATTRIBUTES()
    initialize_object_attributes(obj_attrs, p_name, OBJ_ATTR.OBJ_CASE_INSENSITIVE, root_handle, NULL)

    returned_handle = HANDLE()
    desired_access = ACCESS_MASK.DIRECTORY_QUERY | ACCESS_MASK.DIRECTORY_TRAVERSE
    status = NtOpenDirectoryObject(ctypes.byref(returned_handle), desired_access.value, ctypes.byref(obj_attrs))

    validate_ntstatus(status)

    return returned_handle


def query_directory_object(path_to_dir: str, dir_handle: HANDLE) -> List[NamedObject]:
    """Queries a directory object.

    Parameters:
        path_to_dir: The full path to the specific dir object getting queried
        dir_handle: A pointer to the directory we wish to query.
    """
    context = ULONG(0)
    previous_context = 0

    buffer = ctypes.create_string_buffer(BUFFER_SIZE)
    ntstatus = NtStatusCode.STATUS_MORE_ENTRIES
    objects = []
    while ntstatus == NtStatusCode.STATUS_MORE_ENTRIES:
        return_length = ULONG(0)
        ntstatus = NtQueryDirectoryObject(
            dir_handle,
            ctypes.byref(buffer),
            BUFFER_SIZE,
            False,
            False,
            ctypes.byref(context),
            ctypes.byref(return_length),
        )
        elements_in_buffer = context.value - previous_context
        dir_info_buffer = ctypes.cast(buffer, ctypes.POINTER(OBJECT_DIRECTORY_INFORMATION * elements_in_buffer))

        try:
            validate_ntstatus(ntstatus)
        except NoMoreEntriesError:
            break

        objects += [
            NamedObject.from_directory_information(path_to_dir, directory_information)
            for directory_information in dir_info_buffer.contents
        ]

        previous_context = context.value

    return objects



from unittest.mock import Mock, patch

from acquire.uploaders.plugin_registry import PluginRegistry


def test_registry_functionality():
    data = PluginRegistry[int]("<undefined>")
    data.register("name", int)

    assert data.get("name")(20) == 20


def test_registry_functionality_iterator():
    plugins = [("test", str), ("best", int)]
    data = PluginRegistry("<undefined>", plugins)

    assert len(data.items()) == 2
    assert data.get("best")(20) == 20
    assert data.get("test")("hello") == "hello"


def test_registry_entrypoint():
    mocked_output = Mock()
    with patch.object(PluginRegistry, "_find_entrypoint_data", return_value=[mocked_output]):
        data = PluginRegistry("<undefined>")
        assert data.get(mocked_output.name) == mocked_output.load.return_value


def test_registry_entrypoint_failed():
    mocked_output = Mock()
    mocked_output.load.side_effect = [ModuleNotFoundError]
    data = PluginRegistry("-")

    with patch.object(PluginRegistry, "_find_entrypoint_data", return_value=[mocked_output]):
        data.load_entrypoint_plugins("test")
        assert len(data.items()) == 0



import argparse
import pathlib
import platform
from unittest.mock import MagicMock, patch

import pytest
from dissect.target import Target

from acquire.acquire import MODULES, PROFILES
from acquire.utils import (
    check_and_set_acquire_args,
    check_and_set_log_args,
    create_argument_parser,
    normalize_path,
)


def get_args(**kwargs):
    default_args = {
        "no_log": True,
        "log": None,
        "output": pathlib.Path("."),
        "children": False,
        "upload": None,
        "auto_upload": False,
        "public_key": None,
    }

    parser = create_argument_parser(PROFILES, MODULES)
    default_args = dict(parser.parse_args(args=[])._get_kwargs())
    default_args.update(kwargs)

    args = argparse.Namespace(**default_args)
    return args


def get_mock_path(
    is_dir: bool = True,
    exists: bool = True,
    parent_is_dir: bool = True,
    tree_depth: int = 1,
) -> MagicMock:
    mock_path = MagicMock(spec_set=pathlib.Path)
    mock_path.__str__ = lambda _: "/some/path"

    if exists:
        mock_path.is_dir.return_value = is_dir
        mock_path.is_file.return_value = not is_dir
    else:
        mock_path.is_dir.return_value = False
        mock_path.is_file.return_value = False

    mock_path.exists.return_value = exists

    mock_parent = None
    if tree_depth:
        mock_parent = get_mock_path(is_dir=parent_is_dir, tree_depth=(tree_depth - 1))
    mock_path.parent = mock_parent

    return mock_path


def test_check_and_set_log_args_no_log():
    args = get_args(no_log=True)
    with patch("acquire.utils.get_utc_now_str", return_value="foo"):
        check_and_set_log_args(args)

    assert args.start_time == "foo"
    assert args.log_path is None
    assert not args.log_to_dir
    assert not args.log_delay


@pytest.mark.parametrize(
    "arg_name",
    [
        "log",
        "output",
    ],
)
@pytest.mark.parametrize(
    "mock_path, log_to_dir, log_delay",
    [
        # Log to a directory
        (get_mock_path(), True, True),
        # Log to a file and the file already exists
        (get_mock_path(is_dir=False), False, False),
        # Log to a file which does not yet exist, but it's parent directory does
        (get_mock_path(exists=False), False, False),
    ],
)
def test_check_and_set_log_args(
    arg_name: str,
    mock_path: MagicMock,
    log_to_dir: bool,
    log_delay: bool,
):
    args = get_args(**{arg_name: mock_path})
    with patch("acquire.utils.get_utc_now_str", return_value="foo"):
        check_and_set_log_args(args)

    assert args.start_time == "foo"
    assert args.log_path == mock_path
    assert args.log_to_dir == log_to_dir
    assert args.log_delay == log_delay


def test_check_and_set_log_args_fail_log_to_file_with_children():
    mock_path = get_mock_path(is_dir=False)
    args = get_args(log=mock_path, children=True)
    with pytest.raises(ValueError, match="Log path must be a directory when using --children"):
        check_and_set_log_args(args)


def test_check_and_set_log_args_fail_log_to_path_not_exists():
    mock_path = get_mock_path(exists=False, parent_is_dir=False)
    args = get_args(log=mock_path)
    with pytest.raises(ValueError, match="Log path doesn't exist: /some/path"):
        check_and_set_log_args(args)


@pytest.mark.parametrize(
    "arg_name",
    [
        "upload",
        "auto_upload",
    ],
)
def test_check_and_set_acquire_args_upload_auto_upload(arg_name):
    cagent_key = "bar"
    config = {
        "upload": {"mode": "foo"},
        "cagent_key": cagent_key,
    }

    mock_upload_plugin = MagicMock()
    upload_plugins = {"foo": mock_upload_plugin}

    args = get_args(**{arg_name: True, "config": config})
    check_and_set_acquire_args(args, upload_plugins)

    if arg_name == "upload":
        assert "cagent_key" not in args
    else:
        assert args.cagent_key == cagent_key


@pytest.mark.parametrize(
    "arg_name",
    [
        "upload",
        "auto_upload",
    ],
)
@pytest.mark.parametrize(
    "upload_config, plugin_side_effect, error_match",
    [
        (
            {},
            None,
            "Uploading is not configured",
        ),
        (
            {"upload": {}},
            None,
            "Uploading is not configured",
        ),
        (
            {"upload": {"mode": "bar"}},
            None,
            "Invalid upload mode: bar",
        ),
        (
            {"upload": {"mode": "foo"}},
            ValueError("Plugin init error"),
            "Plugin init error",
        ),
    ],
)
def test_check_and_set_acquire_args_upload_auto_upload_fail(arg_name, upload_config, plugin_side_effect, error_match):
    config = {"cagent_key": "bar"}
    config.update(upload_config)

    mock_upload_plugin = MagicMock()
    if plugin_side_effect:
        mock_upload_plugin.side_effect = plugin_side_effect
    upload_plugins = {"foo": mock_upload_plugin}

    args = get_args(**{arg_name: True, "config": config})

    with pytest.raises(ValueError, match=error_match):
        check_and_set_acquire_args(args, upload_plugins)


@pytest.mark.parametrize(
    "children, output",
    [
        # Output without children to a directory
        (
            False,
            get_mock_path(),
        ),
        # Output with children to a directoy
        (
            True,
            get_mock_path(),
        ),
        # Output without children to a file
        (
            False,
            get_mock_path(is_dir=False),
        ),
        # Output without children to a non-existing file, but the parent directory exists
        (
            False,
            get_mock_path(exists=False),
        ),
    ],
)
def test_check_and_set_acquire_args_output(children, output):
    args = get_args(children=children, output=output, config={})
    result = check_and_set_acquire_args(args, MagicMock())

    assert result is None


@pytest.mark.parametrize(
    "children, output, error_match",
    [
        # Output with children to a file
        (
            True,
            get_mock_path(is_dir=False),
            "Output path must be a directory when using --children",
        ),
        # Output without children to a non-existing file with non-existing parent
        (
            False,
            get_mock_path(exists=False, parent_is_dir=False),
            "Output path doesn't exist: /some/path",
        ),
    ],
)
def test_check_and_set_acquire_args_output_fail(children, output, error_match):
    args = get_args(children=children, output=output, config={})

    with pytest.raises(ValueError, match=error_match):
        check_and_set_acquire_args(args, MagicMock())


def test_check_and_set_acquire_args_encrypt_with_public_key_config():
    config = {"public_key": "PUBLIC KEY"}

    args = get_args(encrypt=True, config=config)
    check_and_set_acquire_args(args, MagicMock())

    assert args.public_key == "PUBLIC KEY"


def test_check_and_set_acquire_args_encrypt_with_public_key_arg():
    mock_path = get_mock_path(is_dir=False)
    mock_path.read_text = lambda: "PUBLIC KEY"

    args = get_args(encrypt=True, public_key=mock_path, config={})
    check_and_set_acquire_args(args, MagicMock())

    assert args.public_key == "PUBLIC KEY"


@pytest.mark.parametrize(
    "public_key",
    [
        None,
        get_mock_path(),
    ],
)
def test_check_and_set_acquire_args_encrypt_without_public_key_fail(public_key):
    args = get_args(encrypt=True, public_key=public_key, config={})

    with pytest.raises(ValueError, match=r"No public key available \(embedded or argument\)"):
        check_and_set_acquire_args(args, MagicMock())


def test_check_and_set_acquire_args_cagent():
    cagent_key = "KEY"
    cagent_certificate = "CERT"
    config = {
        "cagent_key": cagent_key,
        "cagent_certificate": cagent_certificate,
    }
    args = get_args(config=config)
    check_and_set_acquire_args(args, MagicMock())

    assert args.cagent_key == cagent_key
    assert args.cagent_certificate == cagent_certificate


@pytest.mark.parametrize(
    "path, resolve, norm_path, case_sensitive, os",
    [
        (
            pathlib.Path("/foo/bar"),
            False,
            "/foo/bar",
            True,
            "dummy",
        ),
        (
            pathlib.Path("/foo/BAR"),
            False,
            "/foo/bar",
            False,
            "dummy",
        ),
        (
            pathlib.Path("/foo/BAR"),
            False,
            "/foo/BAR",
            True,
            "dummy",
        ),
        (
            pathlib.Path("/foo/../bar"),
            False,
            "/foo/../bar",
            True,
            "dummy",
        ),
        (
            pathlib.Path("/foo/../foo/bar"),
            True,
            "/foo/bar",
            True,
            "dummy",
        ),
        (
            pathlib.PureWindowsPath("c:\\foo\\bar"),
            False,
            "sysvol/foo/bar",
            False,
            "windows",
        ),
        (
            pathlib.PureWindowsPath("C:\\foo\\bar"),
            False,
            "sysvol/foo/bar",
            False,
            "windows",
        ),
        (
            pathlib.PureWindowsPath("\\??\\C:\\foo\\bar"),
            False,
            "sysvol/foo/bar",
            False,
            "windows",
        ),
        (
            pathlib.PureWindowsPath("\\??\\c:\\foo\\bar"),
            False,
            "sysvol/foo/bar",
            False,
            "windows",
        ),
        (
            pathlib.PureWindowsPath("D:\\foo\\bar"),
            False,
            "d:/foo/bar",
            False,
            "windows",
        ),
    ],
)
@pytest.mark.skipif(platform.system() == "Windows", reason="Compares Posix Paths. Needs to be fixed.")
def test_utils_normalize_path(
    mock_target: Target,
    path: pathlib.Path,
    resolve: bool,
    norm_path: str,
    case_sensitive: bool,
    os: str,
) -> None:
    with patch.object(mock_target, "os", new=os):
        with patch.object(mock_target.fs, "_case_sensitive", new=case_sensitive):
            assert normalize_path(mock_target, path, resolve=resolve) == norm_path



from argparse import Namespace
from typing import List
from unittest.mock import patch

import pytest

from acquire.acquire import (
    CONFIG,
    MODULES,
    PROFILES,
    create_argument_parser,
    parse_acquire_args,
)


@pytest.fixture
def acquire_parser_args(config: List, argument_list: List) -> Namespace:
    CONFIG["arguments"] = config
    with patch("argparse._sys.argv", [""] + argument_list):
        return parse_acquire_args(create_argument_parser(PROFILES, MODULES), config=CONFIG)


@pytest.mark.parametrize("config, argument_list", [([], [])])
def test_no_defaults_in_config(acquire_parser_args):
    assert not acquire_parser_args.force_fallback


@pytest.mark.parametrize("config, argument_list", [(["--force-fallback"], [])])
def test_one_config_default_argument(acquire_parser_args):
    assert acquire_parser_args.force_fallback


@pytest.mark.parametrize("config, argument_list", [(["-f", "test"], ["-f", "best"])])
def test_config_default_argument_override(acquire_parser_args):
    assert acquire_parser_args.file == ["best"]



from __future__ import annotations

import dataclasses
import errno
import logging
import subprocess
import textwrap
from collections import defaultdict
from dataclasses import dataclass
from itertools import groupby
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Optional, Sequence, Type, Union

from dissect.target import Target
from dissect.target.exceptions import (
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
    SymlinkRecursionError,
)
from dissect.target.helpers import fsutil

from acquire.utils import StrEnum, get_formatted_exception, normalize_path

if TYPE_CHECKING:
    from acquire.outputs.base import Output

log = logging.getLogger(__name__)


class Outcome(StrEnum):
    SUCCESS = "success"
    FAILURE = "failure"
    MISSING = "missing"
    EMPTY = "empty"


class ArtifactType(StrEnum):
    FILE = "file"
    SYMLINK = "symlink"
    GLOB = "glob"
    DIR = "dir"
    COMMAND = "command"

    # when outcome is known before path is resolved into a file or directory
    PATH = "path"


@dataclass(frozen=True)
class Record:
    module_name: str
    outcome: Outcome
    artifact_type: ArtifactType
    artifact_value: str
    details: Optional[str] = None


def serialize_path(path: Any) -> str:
    if not isinstance(path, fsutil.TargetPath):
        return str(path)

    if not getattr(path, "_fs", None):
        return str(path)

    # Naive way to serialize TargetPath filesystem's metadata is
    # to rely on uniqueness of `path._fs` object
    fs_id = id(path._fs)
    return f"{path._fs.__fstype__}:{fs_id}:{path}"


@dataclass
class CollectionReport:
    registry: set[Record] = dataclasses.field(default_factory=set)

    seen_paths: set[str] = dataclasses.field(default_factory=set)

    def _register(
        self,
        module_name: str,
        outcome: Outcome,
        artifact_type: ArtifactType,
        artifact_value: Union[str, Path],
        details: Optional[str] = None,
    ) -> None:
        if isinstance(artifact_value, Path):
            artifact_value = serialize_path(artifact_value)

        if artifact_type in (ArtifactType.FILE, ArtifactType.DIR):
            self.seen_paths.add(artifact_value)

        self.registry.add(
            Record(
                module_name=module_name,
                outcome=outcome,
                artifact_type=artifact_type,
                artifact_value=artifact_value,
                details=details,
            )
        )

    def add_file_collected(self, module: str, path: Path) -> None:
        self._register(module, Outcome.SUCCESS, ArtifactType.FILE, path)

    def add_symlink_collected(self, module: str, path: Path) -> None:
        self._register(module, Outcome.SUCCESS, ArtifactType.SYMLINK, path)

    def add_symlink_failed(self, module: str, path: Path) -> None:
        self._register(module, Outcome.FAILURE, ArtifactType.SYMLINK, path)

    def add_file_failed(self, module: str, failed_path: Path) -> None:
        exc = get_formatted_exception()
        self._register(module, Outcome.FAILURE, ArtifactType.FILE, failed_path, exc)

    def add_file_missing(self, module: str, missing_path: Path) -> None:
        self._register(module, Outcome.MISSING, ArtifactType.FILE, missing_path)

    def add_glob_collected(self, module: str, pattern: str) -> None:
        self._register(module, Outcome.SUCCESS, ArtifactType.GLOB, pattern)

    def add_glob_failed(self, module: str, failed_pattern: str) -> None:
        exc = get_formatted_exception()
        self._register(module, Outcome.FAILURE, ArtifactType.GLOB, failed_pattern, exc)

    def add_glob_empty(self, module: str, pattern: str) -> None:
        self._register(module, Outcome.EMPTY, ArtifactType.GLOB, pattern)

    def add_dir_collected(self, module: str, path: Path) -> None:
        self._register(module, Outcome.SUCCESS, ArtifactType.DIR, path)

    def add_dir_failed(self, module: str, failed_path: Path) -> None:
        exc = get_formatted_exception()
        self._register(module, Outcome.FAILURE, ArtifactType.DIR, failed_path, exc)

    def add_dir_missing(self, module: str, missing_path: Path) -> None:
        self._register(module, Outcome.MISSING, ArtifactType.DIR, missing_path)

    def add_path_failed(self, module: str, failed_path: Path) -> None:
        exc = get_formatted_exception()
        self._register(module, Outcome.FAILURE, ArtifactType.PATH, failed_path, exc)

    def add_path_missing(self, module: str, missing_path: Path) -> None:
        self._register(module, Outcome.MISSING, ArtifactType.PATH, missing_path)

    def add_command_collected(self, module: str, command_parts: Sequence[str]) -> None:
        self._register(module, Outcome.SUCCESS, ArtifactType.COMMAND, tuple(command_parts))

    def add_command_failed(self, module: str, command_parts: Sequence[str]) -> None:
        exc = get_formatted_exception()
        self._register(module, Outcome.FAILURE, ArtifactType.COMMAND, tuple(command_parts), exc)

    def get_records_per_module_per_outcome(self, serialize_records=False) -> dict[str, dict[str, list[Record]]]:
        grouped_records = defaultdict(lambda: defaultdict(list))

        # sort records by module name and outcome to prepare for grouping
        sorted_registry = sorted(self.registry, key=lambda rec: (rec.module_name, rec.outcome))

        for module_name, records_per_module in groupby(sorted_registry, lambda r: r.module_name):
            for outcome, records_per_module_outcome in groupby(records_per_module, lambda r: r.outcome):
                if serialize_records:
                    records = map(lambda r: dataclasses.asdict(r), records_per_module_outcome)
                else:
                    records = records_per_module_outcome
                grouped_records[module_name][outcome].extend(records)

        return grouped_records

    def get_counts_per_module_per_outcome(self) -> dict[str, dict[str, int]]:
        records_map = self.get_records_per_module_per_outcome()
        for module, records_per_module in records_map.items():
            for outcome, records_per_module_outcome in records_per_module.items():
                records_map[module][outcome] = len(records_per_module_outcome)
        return records_map

    def was_path_seen(self, path: Path) -> bool:
        return serialize_path(path) in self.seen_paths


class Collector:
    METADATA_BASE = "$metadata$"
    COMMAND_OUTPUT_BASE = f"{METADATA_BASE}/command-output"

    def __init__(self, target: Target, output: Output, base: str = "fs", skip_list: Optional[set] = None):
        self.target = target
        self.output = output
        self.base = base
        self.skip_list = skip_list or set()

        self.report = CollectionReport()
        self.bound_module_name = None

        self.output.init(self.target)

    def __enter__(self) -> Collector:
        return self

    def __exit__(self, *args, **kwargs) -> None:
        self.close()

    def bind(self, module: Type) -> None:
        self.bound_module_name = module.__name__

    def unbind(self) -> None:
        self.bound_module_name = None

    def close(self) -> None:
        self.output.close()

    def _create_output_path(self, path: Path, base: Optional[str] = None) -> str:
        base = base or self.base
        outpath = str(path)

        if base:
            # Make sure that `outpath` is not an abolute path, since
            # `fsutil.join()` (that uses `posixpath.join()`) discards all previous path
            # components if an encountered component is an absolute path.
            outpath = outpath.lstrip("/")
            outpath = fsutil.join(base, outpath)

        return outpath

    def collect(
        self, spec: Iterable, module_name: Optional[str] = None, follow: bool = True, volatile: bool = False
    ) -> None:
        module_name = self.bound_module_name or module_name
        if not module_name:
            raise ValueError("Module name must be provided or Collector needs to be bound to a module")

        for spec_item in spec:
            transform_func = None
            if len(spec_item) == 3:
                artifact_type, value, transform_func = spec_item
            else:
                artifact_type, value = spec_item

            if transform_func is not None:
                values = transform_func(self.target, value)
            else:
                values = [value]

            for value in values:
                if artifact_type in (ArtifactType.FILE, ArtifactType.DIR):
                    self.collect_path(value, module_name=module_name, follow=follow, volatile=volatile)
                elif artifact_type == ArtifactType.SYMLINK:
                    self.collect_symlink(value, module_name=module_name)
                elif artifact_type == ArtifactType.GLOB:
                    self.collect_glob(value, module_name=module_name)
                elif artifact_type == ArtifactType.COMMAND:
                    command_parts, output_filename = value
                    self.collect_command_output(command_parts, output_filename, module_name=module_name)
                else:
                    raise ValueError("Unknown artifact type %s in spec: %s", artifact_type, spec)

    def collect_file(
        self,
        path: Union[str, fsutil.TargetPath],
        size: Optional[int] = None,
        outpath: Optional[str] = None,
        module_name: Optional[str] = None,
        base: Optional[str] = None,
        volatile: bool = False,
    ) -> None:
        module_name = self.bound_module_name or module_name
        if not module_name:
            raise ValueError("Module name must be provided or Collector needs to be bound to a module")

        if not isinstance(path, fsutil.TargetPath):
            path = self.target.fs.path(path)

        if self.report.was_path_seen(path):
            log.info("- Collecting file %s: Skipped (DEDUP)", path)
            return

        outpath = self._create_output_path(outpath or path, base)
        entry = path.get()

        try:
            if volatile:
                self.output.write_volatile(outpath, entry, size)
            else:
                self.output.write_entry(outpath, entry, size)

            self.report.add_file_collected(module_name, path)
            result = "OK"
        except FileNotFoundError:
            self.report.add_file_missing(module_name, path)
            result = "File not found"
        except Exception as exc:
            log.error("Failed to collect file", exc_info=True)
            self.report.add_file_failed(module_name, path)
            result = repr(exc)

        log.info("- Collecting file %s: %s", path, result)

    def collect_symlink(self, path: fsutil.TargetPath, module_name: Optional[str] = None) -> None:
        try:
            outpath = self._create_output_path(path)
            self.output.write_bytes(outpath, b"", path.get(), 0)

            self.report.add_symlink_collected(module_name, path)
            result = "OK"
        except Exception as exc:
            self.report.add_symlink_failed(module_name, path)
            log.error("Failed to collect symlink %s (-> %s)", path, path.readlink(), exc_info=True)
            result = repr(exc)

        log.info("- Collecting symlink %s: %s", path, result)

    def collect_dir(
        self,
        path: Union[str, fsutil.TargetPath],
        seen_paths: Optional[set] = None,
        module_name: Optional[str] = None,
        follow: bool = True,
        volatile: bool = False,
    ) -> None:
        module_name = self.bound_module_name or module_name
        if not module_name:
            raise ValueError("Module name must be provided or Collector needs to be bound to a module")

        if not isinstance(path, fsutil.TargetPath):
            path = self.target.fs.path(path)

        log.info("- Collecting directory %s", path)

        seen_paths = seen_paths or set()
        try:
            resolved = path.resolve()
            if resolved in seen_paths:
                log.debug("Breaking out of symlink loop: path %s linking to %s", path, resolved)
                return
            seen_paths.add(resolved)

            for entry in path.iterdir():
                self.collect_path(
                    entry, seen_paths=seen_paths, module_name=module_name, follow=follow, volatile=volatile
                )

        except OSError as error:
            if error.errno == errno.ENOENT:
                self.report.add_dir_missing(module_name, path)
                log.error("- Directory %s is not found", path)
            elif error.errno == errno.EACCES:
                self.report.add_dir_failed(module_name, path)
                log.error("- Permission denied while accessing directory %s", path)
            else:
                self.report.add_dir_failed(module_name, path)
                log.error("- OSError while collecting directory %s (%s)", path, error)
        except Exception:
            self.report.add_dir_failed(module_name, path)
            log.error("- Failed to collect directory %s", path, exc_info=True)

    def collect_glob(self, pattern: str, module_name: Optional[str] = None) -> None:
        module_name = self.bound_module_name or module_name
        if not module_name:
            raise ValueError("Module name must be provided or Collector needs to be bound to a module")

        log.info("- Collecting glob %s", pattern)
        try:
            glob_is_empty = True
            for entry in self.target.fs.path("/").glob(pattern.lstrip("/")):
                glob_is_empty = False
                self.collect_path(entry, module_name=module_name)
        except Exception:
            log.error("- Failed to collect glob %s", pattern, exc_info=True)
            self.report.add_glob_failed(module_name, pattern)
        else:
            if glob_is_empty:
                self.report.add_glob_empty(module_name, pattern)
            else:
                self.report.add_glob_collected(module_name, pattern)

    def collect_path(
        self,
        path: Union[str, fsutil.TargetPath],
        seen_paths: Optional[set] = None,
        module_name: Optional[str] = None,
        follow: bool = True,
        volatile: bool = False,
    ) -> None:
        module_name = self.bound_module_name or module_name
        if not module_name:
            raise ValueError("Module name must be provided or Collector needs to be bound to a module")

        if not isinstance(path, fsutil.TargetPath):
            path = self.target.fs.path(path)

        if self.skip_list and normalize_path(self.target, path) in self.skip_list:
            self.report.add_path_failed(module_name, path)
            log.error("- Skipping collection of %s, path is on the skip list", path)
            return

        try:
            # If a path does not exist, is_dir(), is_file() and is_symlink() will return False (and not raise an
            # exception), so we need to explicitly trigger an exception for this using path.get().
            path.get()
            is_dir = path.is_dir()
            is_file = path.is_file()
            is_symlink = path.is_symlink()
        except OSError as error:
            if error.errno == errno.ENOENT:
                self.report.add_path_missing(module_name, path)
                log.error("- Path %s is not found", path)
            elif error.errno == errno.EACCES:
                self.report.add_path_failed(module_name, path)
                log.error("- Permission denied while accessing path %s", path)
            else:
                self.report.add_path_failed(module_name, path)
                log.error("- OSError while collecting path %s", path)
            return
        except (FileNotFoundError, NotADirectoryError, NotASymlinkError, SymlinkRecursionError, ValueError):
            self.report.add_path_missing(module_name, path)
            log.error("- Path %s is not found", path)
            return
        except Exception:
            self.report.add_path_failed(module_name, path)
            log.error("- Failed to collect path %s", path, exc_info=True)
            return

        if is_symlink:
            self.collect_symlink(path, module_name=module_name)

            if follow:
                # Follow the symlink, call ourself again with the resolved path
                self.collect_path(
                    path.resolve(), seen_paths=seen_paths, module_name=module_name, follow=follow, volatile=volatile
                )
        elif is_dir:
            self.collect_dir(path, seen_paths=seen_paths, module_name=module_name, follow=follow, volatile=volatile)
        elif is_file:
            self.collect_file(path, module_name=module_name, volatile=volatile)
        else:
            self.report.add_path_failed(module_name, path)
            log.error("- Don't know how to collect %s in module %s", path, module_name)

    def collect_command_output(
        self,
        command_parts: list[str],
        output_filename: str,
        module_name: Optional[str] = None,
    ) -> None:
        module_name = self.bound_module_name or module_name
        if not module_name:
            raise ValueError("Module name must be provided or Collector needs to be bound to a module")

        output_base = fsutil.join(self.base, self.COMMAND_OUTPUT_BASE) if self.base else self.COMMAND_OUTPUT_BASE
        full_output_path = fsutil.join(output_base, output_filename)

        log.info("- Collecting output from command `%s`", " ".join(command_parts))
        try:
            command_output = subprocess.check_output(command_parts, stderr=subprocess.STDOUT, shell=True)
            self.output.write_bytes(full_output_path, command_output)
            self.report.add_command_collected(module_name, command_parts)
        except Exception:
            self.report.add_command_failed(module_name, command_parts)
            log.error("- Failed to collect output from command `%s`", " ".join(command_parts), exc_info=True)
            return

    def write_bytes(self, destination_path: str, data: bytes) -> None:
        self.output.write_bytes(destination_path, data, None)
        self.report.add_file_collected(self.bound_module_name, destination_path)


def get_report_summary(report: CollectionReport) -> str:
    """Create a table-view report summary with success/failure/missing/empty counters per module"""

    record_counts = report.get_counts_per_module_per_outcome()

    if not record_counts:
        return ""

    module_name_max_len = max(len(module_name) for module_name in record_counts.keys())
    # Must be as long as a header
    module_name_max_len = max(len("Module"), module_name_max_len)

    row_template = (
        f"{{module_name: >{module_name_max_len}s}} | "
        "{success_count: >10} | "
        "{failure_count: >10} | "
        "{missing_count: >10} | "
        "{empty_count: >10}"
    )

    header = row_template.format(
        module_name="Module",
        success_count="Success",
        failure_count="Failure",
        missing_count="Missing",
        empty_count="Empty",
    )

    splitter = "-" * len(header)

    rows = [
        row_template.format(
            module_name=module_name,
            success_count=counts.get(Outcome.SUCCESS, ""),
            failure_count=counts.get(Outcome.FAILURE, ""),
            missing_count=counts.get(Outcome.MISSING, ""),
            empty_count=counts.get(Outcome.EMPTY, ""),
        )
        for module_name, counts in record_counts.items()
    ]

    total_counts = defaultdict(int)
    for counts in record_counts.values():
        for count_type, value in counts.items():
            total_counts[count_type] += value

    total_counts_row = row_template.format(
        module_name="Total",
        success_count=total_counts[Outcome.SUCCESS],
        failure_count=total_counts[Outcome.FAILURE],
        missing_count=total_counts[Outcome.MISSING],
        empty_count=total_counts[Outcome.EMPTY],
    )

    return "\n".join(
        [
            splitter,
            header,
            splitter,
            *rows,
            splitter,
            total_counts_row,
            splitter,
        ]
    )


def get_full_formatted_report(report: CollectionReport, record_indent: int = 4) -> str:
    """
    Create a full list of successful / failed / missing / empty artifacts collected,
    broken down by module.
    """

    record_line_template = "{record.outcome}: {record.artifact_type}:{record.artifact_value}"
    blocks = []

    for module_name, records_per_module in report.get_records_per_module_per_outcome().items():
        blocks.append(module_name)
        for _, records_per_module_per_outcome in records_per_module.items():
            record_lines = []
            for record in records_per_module_per_outcome:
                record_lines.append(record_line_template.format(record=record))

            paragraph = "\n".join(record_lines)
            paragraph = textwrap.indent(paragraph, prefix=record_indent * " ")

            blocks.append(paragraph)

    return "\n".join(blocks)



from unittest.mock import patch

import pytest

from acquire.esxi import EsxiMemoryManager


def test_vsish_command():
    memory = EsxiMemoryManager()
    with patch("subprocess.check_output", return_value=b"hello"):
        assert memory._execute_vsish_command(["hello", "world"]) == "hello"


@pytest.mark.parametrize(
    "command_output, expected_result",
    [(b"1", "1"), (b"1 2", "1")],
)
def test_process_number(command_output, expected_result):
    memory = EsxiMemoryManager()
    with patch("subprocess.check_output", return_value=command_output):
        assert memory._get_group_id() == expected_result


def test_process_number_empty():
    memory = EsxiMemoryManager()
    with patch("subprocess.check_output", return_value=b""):
        with pytest.raises(ValueError):
            memory._get_group_id()


def test_get_scheme():
    memory = EsxiMemoryManager()
    with patch("subprocess.check_output", return_value=b"hello:10"):
        assert memory._get_memory_scheme() == {"hello": "10"}



import csv
import ctypes
import gzip
import io
import threading
from logging import Filter, LogRecord, getLogger
from queue import Empty, Queue
from typing import Iterable, Optional

from acquire.dynamic.windows.exceptions import OpenProcessError
from acquire.dynamic.windows.ntdll import (
    NtQueryInformationFile,
    NtQueryObject,
    NtQuerySystemInformation,
    NtStatusCode,
    close_handle,
)
from acquire.dynamic.windows.types import (
    BOOL,
    DWORD,
    FILE_INFORMATION_CLASS,
    HANDLE,
    IO_STATUS_BLOCK,
    OBJECT_INFORMATION_CLASS,
    PHANDLE,
    PUBLIC_OBJECT_TYPE_INFORMATION,
    SYSTEM_HANDLE_INFORMATION_EX,
    SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX,
    SYSTEM_INFORMATION_CLASS,
    ULONG,
    DuplicateHandleFlags,
    ErrorCode,
    FileNameInformationFactory,
    Handle,
    ProcessAccess,
)

log = getLogger(__name__)

OpenProcessToken = ctypes.windll.advapi32.OpenProcessToken
OpenProcessToken.argtypes = [HANDLE, DWORD, PHANDLE]

kernel32 = ctypes.windll.kernel32
OpenProcess = kernel32.OpenProcess
OpenProcess.restype = HANDLE

DuplicateHandle = kernel32.DuplicateHandle
DuplicateHandle.argtypes = [HANDLE, HANDLE, HANDLE, ctypes.POINTER(HANDLE), DWORD, BOOL, DWORD]

GetLastError = kernel32.GetLastError
SetLastError = kernel32.SetLastError
GetCurrentProcessId = kernel32.GetCurrentProcessId


class DuplicateFilter(Filter):
    def __init__(self) -> None:
        super().__init__()
        self.msgs = set()

    def filter(self, record: LogRecord) -> bool:
        msg = record.getMessage()
        if show := msg not in self.msgs:
            self.msgs.add(msg)

        return show


def get_handle_type_info(handle: SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX) -> Optional[str]:
    """Return type of handle.

    Args:
        handle: handle for which to return the type information.

    Raises:
        RuntimeError: Raised when the result of the object query is unknown (other than SUCCESS, LENGTH MISMATCH or
        INVALID).
    """
    public_object_type_information = PUBLIC_OBJECT_TYPE_INFORMATION()
    size = DWORD(ctypes.sizeof(public_object_type_information))
    while True:
        result = NtQueryObject(
            handle,
            OBJECT_INFORMATION_CLASS.ObjectTypeInformation,
            ctypes.byref(public_object_type_information),
            size,
            None,
        )

        if result == NtStatusCode.STATUS_SUCCESS:
            return public_object_type_information.name
        elif result == NtStatusCode.STATUS_INFO_LENGTH_MISMATCH:
            size = DWORD(size.value * 4)
            ctypes.resize(public_object_type_information, size.value)
        elif result == NtStatusCode.STATUS_INVALID_HANDLE:
            return None
        else:
            raise RuntimeError(hex(result))


def open_process(pid: int) -> int:
    """Obtain a handle for the given PID.

    More info: https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocess

    Args:
        pid: integer that represents the process ID.

    Raises:
        OpenProcessError: Raies when the System Idle Process, the System Process or one of the CSRSS processes are tried
        to be opened.
    """
    SetLastError(0)

    h_process = OpenProcess(
        ProcessAccess.PROCESS_DUP_HANDLE,
        False,
        pid,
    )

    error = GetLastError()
    if error in [ErrorCode.ERROR_INVALID_PARAMETER, ErrorCode.ERROR_ACCESS_DENIED]:
        raise OpenProcessError(
            f"Likely tried opening the System Idle Process, the System Process or one of the Client Server Run-Time"
            f"Subsystem (CSRSS) processes [pid: {pid}]"
        )

    # No valid handle could be obtained, display the error code
    if h_process == 0:
        raise OpenProcessError(f"OpenProcess Error: 0x{error:x} [pid: {pid}]")

    return h_process


def _get_file_name_thread(h_file: HANDLE, queue: Queue):
    iob = IO_STATUS_BLOCK()
    file_name_information = FileNameInformationFactory()
    file_name = None

    while True:
        result = NtQueryInformationFile(
            h_file,
            ctypes.byref(iob),
            ctypes.byref(file_name_information),
            ULONG(ctypes.sizeof(file_name_information)),
            FILE_INFORMATION_CLASS.FileNameInformation,
        )

        if result == NtStatusCode.STATUS_BUFFER_OVERFLOW:
            file_name_information = FileNameInformationFactory(file_name_information.FileNameLength)
        elif result == NtStatusCode.STATUS_SUCCESS:
            file_name = file_name_information.FileName
            break
        else:
            # Multiple StatusCodes can be observed. In almost all cases FileNameLength is 0. Breaking for now
            break

    queue.put(file_name)


def get_handle_name(pid: int, handle: SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX) -> Optional[str]:
    """Return handle name."""

    remote = pid != GetCurrentProcessId()

    if remote:
        try:
            h_remote = open_process(pid)
        except OpenProcessError as e:
            log.error(e)
            return None
        try:
            handle = duplicate_handle(h_remote, handle)
        except RuntimeError:
            close_handle(h_remote)
            return None

    # Use threading to try (max a second) to get the handle name, since it might hang
    queue = Queue()
    thread = threading.Thread(target=_get_file_name_thread, args=(handle, queue))
    thread.daemon = True
    thread.start()
    thread.join(1.0)

    result = None
    if not thread.is_alive():
        try:
            result = queue.get_nowait()
        except Empty:
            pass

    return result


def get_handles() -> Iterable[Handle]:
    """Returns all handles of a target."""
    system_handle_information = SYSTEM_HANDLE_INFORMATION_EX()
    size = DWORD(ctypes.sizeof(system_handle_information))
    duplicate_filter = DuplicateFilter()
    log.addFilter(duplicate_filter)

    while True:
        result = NtQuerySystemInformation(
            SYSTEM_INFORMATION_CLASS.SystemExtendedHandleInformation,
            ctypes.byref(system_handle_information),
            size,
            ctypes.byref(size),
        )

        if result == NtStatusCode.STATUS_SUCCESS:
            break
        elif result == NtStatusCode.STATUS_INFO_LENGTH_MISMATCH:
            size = DWORD(size.value * 4)
            ctypes.resize(system_handle_information, size.value)
        else:
            raise RuntimeError(hex(result))

    p_handles = ctypes.cast(
        system_handle_information.Handles,
        ctypes.POINTER(SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX * system_handle_information.NumberOfHandles),
    )
    for handle in p_handles.contents:
        handle_type = get_handle_type_info(handle.HandleValue)
        handle_name = get_handle_name(handle.UniqueProcessId, handle.HandleValue)

        if not handle_name:
            continue

        yield Handle(handle, handle_type, handle_name)
    log.removeFilter(duplicate_filter)


def duplicate_handle(h_process: int, handle: SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX) -> HANDLE:
    """Create duplicate handle.

    When the source handle is in use by another program, one needs to create a duplicate handle in order to have full
    control of that handle. This prevents performing operations on the source handle that might have been closed by
    the other program.
    """
    h_dup = HANDLE()
    SetLastError(0)
    result = DuplicateHandle(
        h_process,
        handle,
        kernel32.GetCurrentProcess(),
        ctypes.byref(h_dup),
        0,
        False,
        DuplicateHandleFlags.DUPLICATE_SAME_ACCESS,
    )
    if result == 0:
        raise RuntimeError()

    return h_dup


def serialize_handles_into_csv(rows: Iterable[Handle], compress: bool = True) -> bytes:
    """Serialize handle data into a csv.

    Serialize provided rows into normal or gzip-compressed CSV, and return a tuple
    containing the result bytes.
    """

    raw_buffer = io.BytesIO()

    if compress:
        buffer = gzip.GzipFile(fileobj=raw_buffer, mode="wb")
    else:
        buffer = raw_buffer

    with io.TextIOWrapper(buffer, encoding="utf-8") as wrapper:
        csv_writer = None
        for i, row in enumerate(rows):
            if i == 0:
                csv_writer = csv.DictWriter(wrapper, fieldnames=row.dictionary.keys())
                csv_writer.writeheader()

            csv_writer.writerow(row.dictionary)

    return raw_buffer.getvalue()



from acquire.outputs.dir import DirectoryOutput
from acquire.outputs.tar import TarOutput

__all__ = [
    "DirectoryOutput",
    "TarOutput",
]

OUTPUTS = {
    "tar": TarOutput,
    "dir": DirectoryOutput,
}



import logging
from importlib import metadata
from typing import Generic, ItemsView, Iterable, List, Optional, TypeVar

from acquire.uploaders.plugin import UploaderPlugin

T = TypeVar("T")

log = logging.getLogger("acquire")


class PluginRegistry(Generic[T]):
    """Maintains a collection of plugins.

    Includes functionality to load plugins from classes defined in entrypoints.
    """

    def __init__(self, name: str, plugins: Optional[Iterable[tuple[str, T]]] = None):
        """Create a plugin registry, with an optional plugin iterable.

        Args:
            name: The name of an entrypoint we want to load plugins from.
            plugins: An iterable that contains plugins that we want to register.
        """
        self.plugins: dict[str, T] = dict()

        plugins = plugins or []
        for plugin_name, klass in plugins:
            self.register(plugin_name, klass)

        self.load_entrypoint_plugins(name)

    def register(self, name: str, plugin: T) -> None:
        """Registers a plugin to the plugins dictionary.

        Args:
            name: The key of the dictionary.
            plugin: The plugin class to register.
        """
        self.plugins.update({name: plugin})

    def remove(self, name: str) -> None:
        """Removes a plugin from the plugins dictionary.

        Args:
            name: The key to remove.
        """
        self.plugins.pop(name)

    def items(self) -> ItemsView[str, T]:
        """Returns all the items inside the ``plugins`` dictionary"""
        return self.plugins.items()

    def get(self, name: str) -> T:
        return self.plugins.get(name)

    def _find_entrypoint_data(self, entry_point_name: str) -> List[metadata.EntryPoint]:
        """Searches through the entrypoints to find specific entry_point names.

        Args:
            entry_point_name: The name to search for.

        Returns:
            A list with entry_points associated with that name."""
        try:
            entrypoint_plugins = metadata.entry_points()[entry_point_name]
        except KeyError:
            entrypoint_plugins = []
        return entrypoint_plugins

    def load_entrypoint_plugins(self, name: str) -> None:
        """Loads all classes defined in the entrypoints that use the specified ``name``.

        Loads the class loaded from the entrypoint with the form: ``<name>=<path>:<class>``
        as <name> <loaded class>

        Args:
            name: The entrypoint to search for.
        """
        class_plugins = self._find_entrypoint_data(name)

        for ep in class_plugins:
            try:
                # Loads the class of the entrypoint, and registers it.
                self.register(ep.name, ep.load())
                log.debug(f"Loaded plugin {ep}")
            except ModuleNotFoundError as e:
                log.exception(msg="Entrypoint module could not be loaded.", exc_info=e)
                pass


UploaderRegistry = PluginRegistry[UploaderPlugin]
"""An PluginRegistry instance that registers ``UploaderPlugin`` plugins"""



import io
import tarfile
from pathlib import Path
from typing import BinaryIO, Optional, Union

from dissect.target.filesystem import FilesystemEntry

from acquire.crypt import EncryptedStream
from acquire.outputs.base import Output


class TarOutput(Output):
    """Tar archive acquire output format. Output can be compressed and/or encrypted.

    Args:
        path: The path to write the tar archive to.
        compress: Whether to compress the tar archive.
        encrypt: Whether to encrypt the tar archive.
        public_key: The RSA public key to encrypt the header with.
    """

    def __init__(
        self,
        path: Path,
        compress: bool = False,
        encrypt: bool = False,
        public_key: Optional[bytes] = None,
    ) -> None:
        ext = ".tar" if ".tar" not in path.suffixes else ""
        mode = "w|" if encrypt else "w:"

        if compress:
            ext += ".gz" if ".gz" not in path.suffixes else ""
            mode += "gz"

        if encrypt:
            ext += ".enc"

        self._fh = None
        self.path = path.with_suffix(path.suffix + ext)

        if encrypt:
            self._fh = EncryptedStream(self.path.open("wb"), public_key)
            self.tar = tarfile.open(fileobj=self._fh, mode=mode)
        else:
            self.tar = tarfile.open(name=self.path, mode=mode)

    def write(
        self,
        output_path: str,
        fh: BinaryIO,
        entry: Optional[Union[FilesystemEntry, Path]],
        size: Optional[int] = None,
    ) -> None:
        """Write a filesystem entry or file-like object to a tar file.

        Args:
            output_path: The path of the entry in the output format.
            fh: The file-like object of the entry to write.
            entry: The optional filesystem entry of the entry to write.
            size: The optional file size in bytes of the entry to write.
        """
        stat = None
        size = size or getattr(fh, "size", None)

        if size is None and fh.seekable():
            offset = fh.tell()
            fh.seek(0, io.SEEK_END)
            size = fh.tell()
            fh.seek(offset)

        info = self.tar.tarinfo()
        info.name = output_path
        info.uname = "root"
        info.gname = "root"
        info.size = size or 0

        if entry:
            if entry.is_symlink():
                info.type = tarfile.SYMTYPE
                info.linkname = entry.readlink()

            stat = entry.lstat()

            if stat:
                info.mtime = stat.st_mtime

        self.tar.addfile(info, fh)

    def close(self) -> None:
        """Closes the tar file."""
        self.tar.close()
        if self._fh:
            self._fh.close()



from pathlib import Path

from acquire.tools.decrypter import find_enc_files


def test_find_non_encrypted_files(tmp_path: Path):
    input_files = [tmp_path / "test", tmp_path / "help"]

    for file in input_files:
        file.touch()

    assert find_enc_files(input_files) == []


def test_find_inside_dir(tmp_path: Path):
    output_path = tmp_path.joinpath("output/for/this/test")
    output_path.mkdir(parents=True)

    rel_path = output_path.relative_to(tmp_path)

    expected_outputs = []

    for directory in rel_path.parents:
        searchable_file = tmp_path / directory / "test.enc"
        searchable_file.touch()
        expected_outputs.append(searchable_file)

    assert sorted(find_enc_files([tmp_path])) == sorted(expected_outputs)



import os
from pathlib import Path
from typing import Any, Optional

from acquire.uploaders.plugin import UploaderPlugin


class MinIO(UploaderPlugin):
    def __init__(self, upload: dict[str, str], **kwargs: dict[str, Any]) -> None:
        """An uploader plugin that uploads files using minio.

        Args:
            upload: Contains credentials and information to use for the transfer.

        Raises:
            ValueError: When the configuration is invalid.
        """

        self.endpoint = upload.get("endpoint")
        self.access_id = upload.get("access_id")
        self.access_key = upload.get("access_key")
        self.bucket_name = upload.get("bucket")

        if not all((self.endpoint, self.access_id, self.access_key, self.bucket_name)):
            raise ValueError("Invalid cloud upload configuration")

    def prepare_client(self, paths: list[Path], proxies: Optional[dict[str, str]] = None) -> Any:
        """Prepares a Minio client used to upload files.

        Args:
            paths: The files to upload.
            proxies: The proxies to use during the upload.

        Raises:
            RuntimeError: When the minio module is not installed.
        """
        try:
            import urllib3
            from minio import Minio
        except ImportError:
            raise RuntimeError("Minio upload module is not available")

        http_client = urllib3.proxy_from_url(proxies["http"]) if proxies else None

        return Minio(self.endpoint, self.access_id, self.access_key, http_client=http_client)

    def upload_file(self, client: Any, path: Path) -> None:
        client.fput_object(self.bucket_name, os.path.basename(path), path)

    def finish(self, client: Any) -> None:
        pass



import logging
from pathlib import Path
from typing import Any, Optional

log = logging.getLogger(__name__)

__all__ = [
    "UploaderPlugin",
    "upload_files_using_uploader",
]

MAX_RETRIES = 4


class UploaderPlugin:
    """Creates a typing definition to which an UploaderPlugin should adhere."""

    def prepare_client(self, paths: list[Path], proxies: Optional[dict[str, str]] = None) -> Any:
        """Prepares a client for the upload."""
        raise NotImplementedError()

    def upload_file(self, client: Any, path: Path) -> None:
        """Uploads a file/path using the ``client``."""
        raise NotImplementedError()

    def finish(self, client: Any) -> None:
        """A cleanup step or anything required to finish the upload."""
        raise NotImplementedError()


def upload_files_using_uploader(
    uploader: UploaderPlugin, paths: list[Path], proxies: Optional[dict[str, str]] = None
) -> None:
    """Uploads the files in ``paths`` to a destination.

    Args:
        uploader: The plugin used to upload files.
        paths: A list of files to upload.
        proxies: Proxies used as an intermediate during an upload.
    """
    paths = [Path(path) if isinstance(path, str) else path for path in paths]
    client = uploader.prepare_client(paths, proxies)

    for path in paths:
        for retry in range(MAX_RETRIES):
            if retry == MAX_RETRIES - 1:
                error_log = ("Upload %s FAILED after too many attempts. Stopping.", path)
            else:
                error_log = ("Upload %s FAILED. See log file for details. Retrying", path)

            try:
                log.info("Uploading %s", path)
                uploader.upload_file(client, path)
                log.info("Uploaded %s", path)
                break
            except Exception:
                log.error(*error_log)
                log.exception("")

    uploader.finish(client)



import platform
import shutil
from pathlib import Path
from typing import BinaryIO, Optional, Union

from dissect.target.filesystem import FilesystemEntry

from acquire.outputs.base import Output


class DirectoryOutput(Output):
    def __init__(self, path: Path, **kwargs):
        self.path = path

    def write(
        self, output_path: str, fh: BinaryIO, entry: Optional[Union[FilesystemEntry, Path]], size: Optional[int] = None
    ) -> None:
        if platform.system() == "Windows":
            output_path = output_path.replace(":", "_")

        out_path = self.path.joinpath(output_path)
        out_dir = out_path.parent
        if not out_dir.exists():
            out_dir.mkdir(parents=True)

        with out_path.open("wb") as fhout:
            shutil.copyfileobj(fh, fhout)

    def close(self) -> None:
        pass
